// CommunicationsQueue.h : header file
//
// Copyright(C) 2009 ubisys technologies GmbH, Duesseldorf, Germany.
// All rights reserved.
//
// www.ubisys.de
// support@ubisys.de
//
// THIS SOURCE CODE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//
// IN NO EVENT SHALL UBISYS TECHNOLOGIES BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOURCE CODE OR THE USE OR
// OTHER DEALINGS IN THE SOURCE CODE.
//
// Part of the ubisys(R) 13.56 MHz RFID USB stick demo application
//
// Provides CCommunicationsQueue, an abstract base class for multi-threaded 
// data exchange via pipes or other serializing communication links


/////////////////////////////////////////////////////////////////////////////
// CCommunicationsQueueClient

class CCommunicationsQueueClient
{
	// Overridables
	public:
		// Appends a log entry related to transmission or reception
		virtual void AppendLog(const CPacket &packet, 
			const bool bTransmission) = 0;
};


/////////////////////////////////////////////////////////////////////////////
// CCommunicationsQueue

class CCommunicationsQueue
{
	// typedefs & enums
	public:
		typedef void *(*LPTHREAD_START_ROUTINE)(void *pArgument);
	
	// Construction
	public:
		CCommunicationsQueue(LPTHREAD_START_ROUTINE pfnReadThread, 
			LPTHREAD_START_ROUTINE pfnWriteThread);

	// Attributes
	public:

	// Operations
	public:
		// Initializes the CCommunicationsQueue object. Must be called after
		// construction, prior to calling any other method.
		bool Initialize(const int nDesiredAccess = O_RDWR);

		// Enables or disables Reception
		bool EnableReception(const bool bEnable = true, 
			const bool bClear = true);

		// Enables or disables transmission
		bool EnableTransmission(const bool bEnable = true, 
			const bool bClear = true);

		// Registers an optional client
		void Register(CCommunicationsQueueClient &client);

		// Locks the receiver and transfers any data received
		// to the caller. Returns the number of bytes actually
		// transfered. Should be called in response to the Reception
		// event being signaled.
		virtual int Receive(CPacket &data);

		// Locks the transmitter and transfers the data to the 
		// comm port's tx buffer. Returns the tx queues current size 
		virtual int Send(const CPacket &data);

	// Overridables
	public:
		// Clears the RX buffer
		virtual bool PurgeReceptionBuffer() = 0;

		// Clears the TX buffer
		virtual bool PurgeTransmissionBuffer() = 0;

	// Implementation
	public:
		virtual ~CCommunicationsQueue();

		// Optional client interface for callbacks
		CCommunicationsQueueClient *m_pClient;

		// This thread carries out the reading operations
		pthread_t m_hReadThread;

		// This thread carries out the writing operations
		pthread_t m_hWriteThread;

		// When this event is signaled, new output is available in tx queue
		pthread_cond_t m_conditionTransmit;

		// When this event is signaled, new input is available in rx queue
		pthread_cond_t m_conditionReceive;

		// This critical section object protects the reading thread
		// from being suspended prematurely. In addition, it protects
		// the input queue.
		pthread_mutex_t m_csr;

		// This critical section object protects the writing thread
		// from being suspended prematurely. In addition, it protects
		// the output queue.
		pthread_mutex_t m_csw;

		// Indicates whether the reading thread is running
		bool m_bReadEnable;

		// Indicates whether the writing thread is running
		bool m_bWriteEnable;

		// Entry point into reading thread
		LPTHREAD_START_ROUTINE m_pfnReadThread;

		// Entry point into writing thread
		LPTHREAD_START_ROUTINE m_pfnWriteThread;
		
        // Set to true by read thread (startup indication) 
        volatile bool m_bReadThreadStarted;
        
        // Set to true by write thread (startup indication) 
        volatile bool m_bWriteThreadStarted;
		
		// Causes reader and writer threads to terminate
		volatile bool m_bTerminate;
	
		// Terminate signal handler
		static void OnSignalTerminate(int nSignal, siginfo_t *pInfo, void *pContext);
				
		// Must be called by each thread
		static void SetupTerminateHandler();

		// Instructs the worker threads to terminate and waits until the
		// both of them have actually terminated
		bool TerminateThreads();

	public:
		// The input queue. Data read from the serial port is stored
		// here, as long as Reception is enabled
		CPacket m_receive;

		// The output queue. Data to be sent over the serial port is
		// stored here. When transmission is enabled, elements are
		// dequeued and transmitted over the serial port.
		CPacket m_transmit;
};

inline void CCommunicationsQueue::Register(CCommunicationsQueueClient &client)
	{ m_pClient = &client; }
