// SerialCommandInterface.h : header file
//
// Copyright(C) 2009 ubisys technologies GmbH, Duesseldorf, Germany.
// All rights reserved.
//
// www.ubisys.de
// support@ubisys.de
//
// THIS SOURCE CODE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//
// IN NO EVENT SHALL UBISYS TECHNOLOGIES BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOURCE CODE OR THE USE OR
// OTHER DEALINGS IN THE SOURCE CODE.
//
// Part of the ubisys(R) 13.56 MHz RFID USB stick demo application
//
// Provides simple high-level functions to control a ubisys RFID USB module
// via its AT like command set

/////////////////////////////////////////////////////////////////////////////
// CSerialCommandInterface

class CSerialCommandInterface
{
	// typedefs & enums
	public:
		// Maximum response frame size
		enum { responseCapacity = 128 };

		typedef bool(*RECEIVECALLBACK)(CSerialCommandInterface &sci,
			std::vector<std::string> &lines, void *pArgument);

		// Info flags used by get system information command
		enum { infoDSFID = 0x01, infoAFI = 0x02, infoVICCMemorySize = 0x04, 
			infoICReference = 0x08 };

	// Construction
	public:
		CSerialCommandInterface(CCommQueue &queue);

	// Attributes
	public:
		// Communications queue that is used to exchange SCI frames. This can 
		// be a standard serial port, a USB connection or whatever physical
		// layer that implements the CCommunicationQueue interface
		CCommQueue &m_queue;

	// Operations
	public:
		// Can be used when ReceiveResponse() or ReceiveResponseEx() have 
		// been called with long timeouts to force the reception routine to
		// terminate immediately
		void AbortReception();
				
		// Enables or disable command echo
		bool EnableEcho(const bool bEnable = true);

		// Returns model string, firmware version and build date, and 
		// product serial number
		bool GetVersion(std::string &strVersion, std::string &strSerialNumber);

		// Makes the device enter its flash update mode. Once the command
		// has been issued, the device does no longer operate normally.
		// It disconnects and appears as a different device
		bool PrepareFlashUpdate();

		// Enables the RF field
		bool EnableField(const bool bEnable = true);

		// Determines current RF field status
		bool QueryField(bool &bEnabled);

		// Enters or leaves automatic scan mode
		bool EnableAutoScan(const bool bEnable = true);

		// Call this method after scan mode has been enabled to anticipate
		// automatic inventory scan results
		bool AnticipateScanResults(const RECEIVECALLBACK pfnReceive, 
			void *const pArgument = 0, const unsigned int nTimeout = 0);

		// Performs an inventory scan
		bool Inventory(std::list<unsigned long long> &uids);

		// Note: Information other that UID is valid only, if related bit is
		// set in bFlags
		bool GetSystemInformation(unsigned long long &nUID, 
			unsigned char &bFlags, unsigned char &bDSFID, unsigned char &bAFI, 
			unsigned char &bBlockCount, unsigned char &bBlockSize,
			unsigned char &bICReference);

		// Reads a block of memory using the read single block command
		bool ReadSingleBlock(const unsigned char nBlock, CPacket &data);

		// Writes a block of memory using the write single block command.
		// Supports both ISO 15693 methods of writing according to the 
		// specified option flag
		bool WriteSingleBlock(const unsigned char nBlock, 
			const CPacket &data, const bool bOptionFlag = false);

	// Implementation
	protected:
		// Forces ReceiveResponse() to terminate without timeout
		volatile bool m_bAbort;
	
		// Transmits the specified request frame and returns immediately. 
		// Does not wait for a response
		bool SendRequest(CPacket &request);

		// Tries to receive data for at most dwTimeout milliseconds. Returns
		// false when the timeout has elapsed and there was not enough data,
		// true when the specified amount of data has been collected
		bool ReceiveResponse(CPacket &response,
			const unsigned int nTimeout, bool *const pbAborted = 0);

		// Verifies a response frame. Checks if the actual length matches 
		// the expected length, and checks for error flags via GetStatus().
		// For response frames that contain payload, checks if the actual
		// type matches the expected type 
		static bool VerifyResponse(const std::vector<std::string> &response);

		// Transmits the specified frame and waits for the mating response 
		// frame. Returns true upon success, false upon error. The status 
		// is returned, and also stored in m_bStatus. A timeout may be 
		// specified in milliseconds, or INFINITE for no timeout
		bool ExchangeCommand(CPacket &request, CPacket &response, 
			const unsigned int nTimeout = 500);

		// Runs ExchangeCommand() and VerifyResponse() with the supplied 
		// arguments
		bool ExecuteEx(CPacket &request, std::vector<std::string> *pResponse = 0, 
			const unsigned int nTimeout = 500, CPacket *pRawResponse = 0);

		// Facilitates execution of simple SCI commands by accepting an
		// ASCII command string, which is appended a carriage-return and 
		// running ExecuteEx on the resulting request
		bool Execute(const char *const pszRequest, std::vector<std::string> *pResponse = 0, 
			const unsigned int dwTimeout = 500, CPacket *pRawResponse = 0);

		// Splits a response frame into an array of strings
		static bool SplitResponse(const CPacket &response, 
			std::vector<std::string> &lines);

		// Remove leading white space (MFC replica)
		static std::string TrimLeft(const std::string &str,
			const char *const pszDelimiters);

		// Remove trailing white space (MFC replica)
		static std::string TrimRight(const std::string &str,
			const char *const pszDelimiters);

		// Remove leading and trailing white space (MFC replica)
		static std::string Trim(const std::string &str,
			const char *const pszDelimiters);
			
		// Returns span of characters (MFC replica) 
		static std::string SpanExcluding(const std::string &str,
			const char *const pszDelimiters);			

		// Tokenization function (MFC replica)
		static std::string Tokenize(const std::string &str, 
			const char *const pszDelimiters, std::string::size_type &start);

	public:
		// Converts an ASCII hex representation of a unique ID to an unsigned
		// 64-bit value
		static bool RetrieveUniqueID(const std::string &strUniqueID,
			unsigned long long &nUniqueID);
};
